# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.comment import Comment
from ..types.error_info import ErrorInfo
from ..types.error_info_write import ErrorInfoWrite
from ..types.feedback_score_batch_item import FeedbackScoreBatchItem
from ..types.feedback_score_names_public import FeedbackScoreNamesPublic
from ..types.feedback_score_source import FeedbackScoreSource
from ..types.json_list_string import JsonListString
from ..types.json_list_string_write import JsonListStringWrite
from ..types.project_stats_public import ProjectStatsPublic
from ..types.span_filter_public import SpanFilterPublic
from ..types.span_page_public import SpanPagePublic
from ..types.span_public import SpanPublic
from ..types.span_update import SpanUpdate
from ..types.span_update_source import SpanUpdateSource
from ..types.span_update_type import SpanUpdateType
from ..types.span_write import SpanWrite
from ..types.span_write_source import SpanWriteSource
from ..types.span_write_type import SpanWriteType
from ..types.value_entry import ValueEntry
from .raw_client import AsyncRawSpansClient, RawSpansClient
from .types.find_feedback_score_names1request_type import FindFeedbackScoreNames1RequestType
from .types.get_span_stats_request_type import GetSpanStatsRequestType
from .types.get_spans_by_project_request_type import GetSpansByProjectRequestType
from .types.span_search_stream_request_public_type import SpanSearchStreamRequestPublicType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class SpansClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawSpansClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawSpansClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawSpansClient
        """
        return self._raw_client

    def add_span_comment(
        self,
        id_: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Add span comment

        Parameters
        ----------
        id_ : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.add_span_comment(id_='id', text='text', )
        """
        _response = self._raw_client.add_span_comment(
            id_,
            text=text,
            id=id,
            created_at=created_at,
            last_updated_at=last_updated_at,
            created_by=created_by,
            last_updated_by=last_updated_by,
            request_options=request_options,
        )
        return _response.data

    def add_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        value: float,
        source: FeedbackScoreSource,
        category_name: typing.Optional[str] = OMIT,
        reason: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        value_by_author: typing.Optional[typing.Dict[str, ValueEntry]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Add span feedback score

        Parameters
        ----------
        id : str

        name : str

        value : float

        source : FeedbackScoreSource

        category_name : typing.Optional[str]

        reason : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        value_by_author : typing.Optional[typing.Dict[str, ValueEntry]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.add_span_feedback_score(id='id', name='name', value=1.1, source="ui", )
        """
        _response = self._raw_client.add_span_feedback_score(
            id,
            name=name,
            value=value,
            source=source,
            category_name=category_name,
            reason=reason,
            created_at=created_at,
            last_updated_at=last_updated_at,
            created_by=created_by,
            last_updated_by=last_updated_by,
            value_by_author=value_by_author,
            request_options=request_options,
        )
        return _response.data

    def create_spans(
        self, *, spans: typing.Sequence[SpanWrite], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Create spans

        Parameters
        ----------
        spans : typing.Sequence[SpanWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import SpanWrite
        import datetime
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.create_spans(spans=[SpanWrite(start_time=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), )], )
        """
        _response = self._raw_client.create_spans(spans=spans, request_options=request_options)
        return _response.data

    def batch_update_spans(
        self,
        *,
        ids: typing.Sequence[str],
        update: SpanUpdate,
        merge_tags: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update multiple spans

        Parameters
        ----------
        ids : typing.Sequence[str]
            List of span IDs to update (max 1000)

        update : SpanUpdate

        merge_tags : typing.Optional[bool]
            If true, merge tags with existing tags instead of replacing them. Default: false

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import SpanUpdate
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.batch_update_spans(ids=['ids'], update=SpanUpdate(trace_id='trace_id', ), )
        """
        _response = self._raw_client.batch_update_spans(
            ids=ids, update=update, merge_tags=merge_tags, request_options=request_options
        )
        return _response.data

    def get_spans_by_project(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        project_name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpansByProjectRequestType] = None,
        filters: typing.Optional[str] = None,
        truncate: typing.Optional[bool] = None,
        strip_attachments: typing.Optional[bool] = None,
        sorting: typing.Optional[str] = None,
        exclude: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SpanPagePublic:
        """
        Get spans by project_name or project_id and optionally by trace_id and/or type

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpansByProjectRequestType]

        filters : typing.Optional[str]

        truncate : typing.Optional[bool]

        strip_attachments : typing.Optional[bool]

        sorting : typing.Optional[str]

        exclude : typing.Optional[str]

        search : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SpanPagePublic
            Spans resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.get_spans_by_project()
        """
        _response = self._raw_client.get_spans_by_project(
            page=page,
            size=size,
            project_name=project_name,
            project_id=project_id,
            trace_id=trace_id,
            type=type,
            filters=filters,
            truncate=truncate,
            strip_attachments=strip_attachments,
            sorting=sorting,
            exclude=exclude,
            search=search,
            from_time=from_time,
            to_time=to_time,
            request_options=request_options,
        )
        return _response.data

    def create_span(
        self,
        *,
        start_time: dt.datetime,
        id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanWriteType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListStringWrite] = OMIT,
        output: typing.Optional[JsonListStringWrite] = OMIT,
        metadata: typing.Optional[JsonListStringWrite] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        error_info: typing.Optional[ErrorInfoWrite] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        total_estimated_cost_version: typing.Optional[str] = OMIT,
        ttft: typing.Optional[float] = OMIT,
        source: typing.Optional[SpanWriteSource] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create span

        Parameters
        ----------
        start_time : dt.datetime

        id : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        trace_id : typing.Optional[str]

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanWriteType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListStringWrite]

        output : typing.Optional[JsonListStringWrite]

        metadata : typing.Optional[JsonListStringWrite]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        usage : typing.Optional[typing.Dict[str, int]]

        error_info : typing.Optional[ErrorInfoWrite]

        last_updated_at : typing.Optional[dt.datetime]

        total_estimated_cost : typing.Optional[float]

        total_estimated_cost_version : typing.Optional[str]

        ttft : typing.Optional[float]
            Time to first token in milliseconds

        source : typing.Optional[SpanWriteSource]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        import datetime
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.create_span(start_time=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), )
        """
        _response = self._raw_client.create_span(
            start_time=start_time,
            id=id,
            project_name=project_name,
            trace_id=trace_id,
            parent_span_id=parent_span_id,
            name=name,
            type=type,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            model=model,
            provider=provider,
            tags=tags,
            usage=usage,
            error_info=error_info,
            last_updated_at=last_updated_at,
            total_estimated_cost=total_estimated_cost,
            total_estimated_cost_version=total_estimated_cost_version,
            ttft=ttft,
            source=source,
            request_options=request_options,
        )
        return _response.data

    def get_span_by_id(
        self,
        id: str,
        *,
        strip_attachments: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SpanPublic:
        """
        Get span by id

        Parameters
        ----------
        id : str

        strip_attachments : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SpanPublic
            Span resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.get_span_by_id(id='id', )
        """
        _response = self._raw_client.get_span_by_id(
            id, strip_attachments=strip_attachments, request_options=request_options
        )
        return _response.data

    def delete_span_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete span by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.delete_span_by_id(id='id', )
        """
        _response = self._raw_client.delete_span_by_id(id, request_options=request_options)
        return _response.data

    def update_span(
        self,
        id: str,
        *,
        trace_id: str,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanUpdateType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListString] = OMIT,
        output: typing.Optional[JsonListString] = OMIT,
        metadata: typing.Optional[JsonListString] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        tags_to_add: typing.Optional[typing.Sequence[str]] = OMIT,
        tags_to_remove: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        error_info: typing.Optional[ErrorInfo] = OMIT,
        ttft: typing.Optional[float] = OMIT,
        source: typing.Optional[SpanUpdateSource] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update span by id

        Parameters
        ----------
        id : str

        trace_id : str

        project_name : typing.Optional[str]
            If null and project_id not specified, Default Project is assumed

        project_id : typing.Optional[str]
            If null and project_name not specified, Default Project is assumed

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanUpdateType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListString]

        output : typing.Optional[JsonListString]

        metadata : typing.Optional[JsonListString]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]
            Tags

        tags_to_add : typing.Optional[typing.Sequence[str]]
            Tags to add

        tags_to_remove : typing.Optional[typing.Sequence[str]]
            Tags to remove

        usage : typing.Optional[typing.Dict[str, int]]

        total_estimated_cost : typing.Optional[float]

        error_info : typing.Optional[ErrorInfo]

        ttft : typing.Optional[float]

        source : typing.Optional[SpanUpdateSource]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.update_span(id='id', trace_id='trace_id', )
        """
        _response = self._raw_client.update_span(
            id,
            trace_id=trace_id,
            project_name=project_name,
            project_id=project_id,
            parent_span_id=parent_span_id,
            name=name,
            type=type,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            model=model,
            provider=provider,
            tags=tags,
            tags_to_add=tags_to_add,
            tags_to_remove=tags_to_remove,
            usage=usage,
            total_estimated_cost=total_estimated_cost,
            error_info=error_info,
            ttft=ttft,
            source=source,
            request_options=request_options,
        )
        return _response.data

    def delete_span_comments(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete span comments

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.delete_span_comments(ids=['ids'], )
        """
        _response = self._raw_client.delete_span_comments(ids=ids, request_options=request_options)
        return _response.data

    def delete_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        author: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Delete span feedback score

        Parameters
        ----------
        id : str

        name : str

        author : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.delete_span_feedback_score(id='id', name='name', )
        """
        _response = self._raw_client.delete_span_feedback_score(
            id, name=name, author=author, request_options=request_options
        )
        return _response.data

    def find_feedback_score_names1(
        self,
        *,
        project_id: typing.Optional[str] = None,
        type: typing.Optional[FindFeedbackScoreNames1RequestType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> FeedbackScoreNamesPublic:
        """
        Find Feedback Score names

        Parameters
        ----------
        project_id : typing.Optional[str]

        type : typing.Optional[FindFeedbackScoreNames1RequestType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackScoreNamesPublic
            Feedback Scores resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.find_feedback_score_names1()
        """
        _response = self._raw_client.find_feedback_score_names1(
            project_id=project_id, type=type, request_options=request_options
        )
        return _response.data

    def get_span_comment(
        self, span_id: str, comment_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> Comment:
        """
        Get span comment

        Parameters
        ----------
        span_id : str

        comment_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Comment
            Comment resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.get_span_comment(span_id='spanId', comment_id='commentId', )
        """
        _response = self._raw_client.get_span_comment(span_id, comment_id, request_options=request_options)
        return _response.data

    def get_span_stats(
        self,
        *,
        project_id: typing.Optional[str] = None,
        project_name: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpanStatsRequestType] = None,
        filters: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectStatsPublic:
        """
        Get span stats

        Parameters
        ----------
        project_id : typing.Optional[str]

        project_name : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpanStatsRequestType]

        filters : typing.Optional[str]

        search : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectStatsPublic
            Span stats resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.get_span_stats()
        """
        _response = self._raw_client.get_span_stats(
            project_id=project_id,
            project_name=project_name,
            trace_id=trace_id,
            type=type,
            filters=filters,
            search=search,
            from_time=from_time,
            to_time=to_time,
            request_options=request_options,
        )
        return _response.data

    def score_batch_of_spans(
        self,
        *,
        scores: typing.Sequence[FeedbackScoreBatchItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Batch feedback scoring for spans

        Parameters
        ----------
        scores : typing.Sequence[FeedbackScoreBatchItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import FeedbackScoreBatchItem
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.score_batch_of_spans(scores=[FeedbackScoreBatchItem(name='name', value=1.1, source="ui", id='id', )], )
        """
        _response = self._raw_client.score_batch_of_spans(scores=scores, request_options=request_options)
        return _response.data

    def search_spans(
        self,
        *,
        trace_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanSearchStreamRequestPublicType] = OMIT,
        filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        truncate: typing.Optional[bool] = OMIT,
        from_time: typing.Optional[dt.datetime] = OMIT,
        to_time: typing.Optional[dt.datetime] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[bytes]:
        """
        Search spans

        Parameters
        ----------
        trace_id : typing.Optional[str]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        type : typing.Optional[SpanSearchStreamRequestPublicType]

        filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        limit : typing.Optional[int]
            Max number of spans to be streamed

        last_retrieved_id : typing.Optional[str]

        truncate : typing.Optional[bool]
            Truncate image included in either input, output or metadata

        from_time : typing.Optional[dt.datetime]
            Filter spans created from this time (ISO-8601 format).

        to_time : typing.Optional[dt.datetime]
            Filter spans created up to this time (ISO-8601 format). If not provided, defaults to current time. Must be after 'from_time'.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[bytes]
            Spans stream or error during process

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.search_spans()
        """
        with self._raw_client.search_spans(
            trace_id=trace_id,
            project_name=project_name,
            project_id=project_id,
            type=type,
            filters=filters,
            limit=limit,
            last_retrieved_id=last_retrieved_id,
            truncate=truncate,
            from_time=from_time,
            to_time=to_time,
            request_options=request_options,
        ) as r:
            yield from r.data

    def update_span_comment(
        self,
        comment_id: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update span comment by id

        Parameters
        ----------
        comment_id : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.spans.update_span_comment(comment_id='commentId', text='text', )
        """
        _response = self._raw_client.update_span_comment(
            comment_id,
            text=text,
            id=id,
            created_at=created_at,
            last_updated_at=last_updated_at,
            created_by=created_by,
            last_updated_by=last_updated_by,
            request_options=request_options,
        )
        return _response.data


class AsyncSpansClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawSpansClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawSpansClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawSpansClient
        """
        return self._raw_client

    async def add_span_comment(
        self,
        id_: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Add span comment

        Parameters
        ----------
        id_ : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.add_span_comment(id_='id', text='text', )
        asyncio.run(main())
        """
        _response = await self._raw_client.add_span_comment(
            id_,
            text=text,
            id=id,
            created_at=created_at,
            last_updated_at=last_updated_at,
            created_by=created_by,
            last_updated_by=last_updated_by,
            request_options=request_options,
        )
        return _response.data

    async def add_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        value: float,
        source: FeedbackScoreSource,
        category_name: typing.Optional[str] = OMIT,
        reason: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        value_by_author: typing.Optional[typing.Dict[str, ValueEntry]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Add span feedback score

        Parameters
        ----------
        id : str

        name : str

        value : float

        source : FeedbackScoreSource

        category_name : typing.Optional[str]

        reason : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        value_by_author : typing.Optional[typing.Dict[str, ValueEntry]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.add_span_feedback_score(id='id', name='name', value=1.1, source="ui", )
        asyncio.run(main())
        """
        _response = await self._raw_client.add_span_feedback_score(
            id,
            name=name,
            value=value,
            source=source,
            category_name=category_name,
            reason=reason,
            created_at=created_at,
            last_updated_at=last_updated_at,
            created_by=created_by,
            last_updated_by=last_updated_by,
            value_by_author=value_by_author,
            request_options=request_options,
        )
        return _response.data

    async def create_spans(
        self, *, spans: typing.Sequence[SpanWrite], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Create spans

        Parameters
        ----------
        spans : typing.Sequence[SpanWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import SpanWrite
        import datetime
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.create_spans(spans=[SpanWrite(start_time=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), )], )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_spans(spans=spans, request_options=request_options)
        return _response.data

    async def batch_update_spans(
        self,
        *,
        ids: typing.Sequence[str],
        update: SpanUpdate,
        merge_tags: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update multiple spans

        Parameters
        ----------
        ids : typing.Sequence[str]
            List of span IDs to update (max 1000)

        update : SpanUpdate

        merge_tags : typing.Optional[bool]
            If true, merge tags with existing tags instead of replacing them. Default: false

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import SpanUpdate
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.batch_update_spans(ids=['ids'], update=SpanUpdate(trace_id='trace_id', ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.batch_update_spans(
            ids=ids, update=update, merge_tags=merge_tags, request_options=request_options
        )
        return _response.data

    async def get_spans_by_project(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        project_name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpansByProjectRequestType] = None,
        filters: typing.Optional[str] = None,
        truncate: typing.Optional[bool] = None,
        strip_attachments: typing.Optional[bool] = None,
        sorting: typing.Optional[str] = None,
        exclude: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SpanPagePublic:
        """
        Get spans by project_name or project_id and optionally by trace_id and/or type

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpansByProjectRequestType]

        filters : typing.Optional[str]

        truncate : typing.Optional[bool]

        strip_attachments : typing.Optional[bool]

        sorting : typing.Optional[str]

        exclude : typing.Optional[str]

        search : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SpanPagePublic
            Spans resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.get_spans_by_project()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_spans_by_project(
            page=page,
            size=size,
            project_name=project_name,
            project_id=project_id,
            trace_id=trace_id,
            type=type,
            filters=filters,
            truncate=truncate,
            strip_attachments=strip_attachments,
            sorting=sorting,
            exclude=exclude,
            search=search,
            from_time=from_time,
            to_time=to_time,
            request_options=request_options,
        )
        return _response.data

    async def create_span(
        self,
        *,
        start_time: dt.datetime,
        id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanWriteType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListStringWrite] = OMIT,
        output: typing.Optional[JsonListStringWrite] = OMIT,
        metadata: typing.Optional[JsonListStringWrite] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        error_info: typing.Optional[ErrorInfoWrite] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        total_estimated_cost_version: typing.Optional[str] = OMIT,
        ttft: typing.Optional[float] = OMIT,
        source: typing.Optional[SpanWriteSource] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create span

        Parameters
        ----------
        start_time : dt.datetime

        id : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        trace_id : typing.Optional[str]

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanWriteType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListStringWrite]

        output : typing.Optional[JsonListStringWrite]

        metadata : typing.Optional[JsonListStringWrite]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        usage : typing.Optional[typing.Dict[str, int]]

        error_info : typing.Optional[ErrorInfoWrite]

        last_updated_at : typing.Optional[dt.datetime]

        total_estimated_cost : typing.Optional[float]

        total_estimated_cost_version : typing.Optional[str]

        ttft : typing.Optional[float]
            Time to first token in milliseconds

        source : typing.Optional[SpanWriteSource]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import datetime
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.create_span(start_time=datetime.datetime.fromisoformat("2024-01-15 09:30:00+00:00", ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_span(
            start_time=start_time,
            id=id,
            project_name=project_name,
            trace_id=trace_id,
            parent_span_id=parent_span_id,
            name=name,
            type=type,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            model=model,
            provider=provider,
            tags=tags,
            usage=usage,
            error_info=error_info,
            last_updated_at=last_updated_at,
            total_estimated_cost=total_estimated_cost,
            total_estimated_cost_version=total_estimated_cost_version,
            ttft=ttft,
            source=source,
            request_options=request_options,
        )
        return _response.data

    async def get_span_by_id(
        self,
        id: str,
        *,
        strip_attachments: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SpanPublic:
        """
        Get span by id

        Parameters
        ----------
        id : str

        strip_attachments : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SpanPublic
            Span resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.get_span_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_span_by_id(
            id, strip_attachments=strip_attachments, request_options=request_options
        )
        return _response.data

    async def delete_span_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete span by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.delete_span_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_span_by_id(id, request_options=request_options)
        return _response.data

    async def update_span(
        self,
        id: str,
        *,
        trace_id: str,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanUpdateType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListString] = OMIT,
        output: typing.Optional[JsonListString] = OMIT,
        metadata: typing.Optional[JsonListString] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        tags_to_add: typing.Optional[typing.Sequence[str]] = OMIT,
        tags_to_remove: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        error_info: typing.Optional[ErrorInfo] = OMIT,
        ttft: typing.Optional[float] = OMIT,
        source: typing.Optional[SpanUpdateSource] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update span by id

        Parameters
        ----------
        id : str

        trace_id : str

        project_name : typing.Optional[str]
            If null and project_id not specified, Default Project is assumed

        project_id : typing.Optional[str]
            If null and project_name not specified, Default Project is assumed

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanUpdateType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListString]

        output : typing.Optional[JsonListString]

        metadata : typing.Optional[JsonListString]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]
            Tags

        tags_to_add : typing.Optional[typing.Sequence[str]]
            Tags to add

        tags_to_remove : typing.Optional[typing.Sequence[str]]
            Tags to remove

        usage : typing.Optional[typing.Dict[str, int]]

        total_estimated_cost : typing.Optional[float]

        error_info : typing.Optional[ErrorInfo]

        ttft : typing.Optional[float]

        source : typing.Optional[SpanUpdateSource]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.update_span(id='id', trace_id='trace_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_span(
            id,
            trace_id=trace_id,
            project_name=project_name,
            project_id=project_id,
            parent_span_id=parent_span_id,
            name=name,
            type=type,
            end_time=end_time,
            input=input,
            output=output,
            metadata=metadata,
            model=model,
            provider=provider,
            tags=tags,
            tags_to_add=tags_to_add,
            tags_to_remove=tags_to_remove,
            usage=usage,
            total_estimated_cost=total_estimated_cost,
            error_info=error_info,
            ttft=ttft,
            source=source,
            request_options=request_options,
        )
        return _response.data

    async def delete_span_comments(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete span comments

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.delete_span_comments(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_span_comments(ids=ids, request_options=request_options)
        return _response.data

    async def delete_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        author: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Delete span feedback score

        Parameters
        ----------
        id : str

        name : str

        author : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.delete_span_feedback_score(id='id', name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_span_feedback_score(
            id, name=name, author=author, request_options=request_options
        )
        return _response.data

    async def find_feedback_score_names1(
        self,
        *,
        project_id: typing.Optional[str] = None,
        type: typing.Optional[FindFeedbackScoreNames1RequestType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> FeedbackScoreNamesPublic:
        """
        Find Feedback Score names

        Parameters
        ----------
        project_id : typing.Optional[str]

        type : typing.Optional[FindFeedbackScoreNames1RequestType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackScoreNamesPublic
            Feedback Scores resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.find_feedback_score_names1()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_feedback_score_names1(
            project_id=project_id, type=type, request_options=request_options
        )
        return _response.data

    async def get_span_comment(
        self, span_id: str, comment_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> Comment:
        """
        Get span comment

        Parameters
        ----------
        span_id : str

        comment_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Comment
            Comment resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.get_span_comment(span_id='spanId', comment_id='commentId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_span_comment(span_id, comment_id, request_options=request_options)
        return _response.data

    async def get_span_stats(
        self,
        *,
        project_id: typing.Optional[str] = None,
        project_name: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpanStatsRequestType] = None,
        filters: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectStatsPublic:
        """
        Get span stats

        Parameters
        ----------
        project_id : typing.Optional[str]

        project_name : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpanStatsRequestType]

        filters : typing.Optional[str]

        search : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectStatsPublic
            Span stats resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.get_span_stats()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_span_stats(
            project_id=project_id,
            project_name=project_name,
            trace_id=trace_id,
            type=type,
            filters=filters,
            search=search,
            from_time=from_time,
            to_time=to_time,
            request_options=request_options,
        )
        return _response.data

    async def score_batch_of_spans(
        self,
        *,
        scores: typing.Sequence[FeedbackScoreBatchItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Batch feedback scoring for spans

        Parameters
        ----------
        scores : typing.Sequence[FeedbackScoreBatchItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import FeedbackScoreBatchItem
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.score_batch_of_spans(scores=[FeedbackScoreBatchItem(name='name', value=1.1, source="ui", id='id', )], )
        asyncio.run(main())
        """
        _response = await self._raw_client.score_batch_of_spans(scores=scores, request_options=request_options)
        return _response.data

    async def search_spans(
        self,
        *,
        trace_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanSearchStreamRequestPublicType] = OMIT,
        filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        truncate: typing.Optional[bool] = OMIT,
        from_time: typing.Optional[dt.datetime] = OMIT,
        to_time: typing.Optional[dt.datetime] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[bytes]:
        """
        Search spans

        Parameters
        ----------
        trace_id : typing.Optional[str]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        type : typing.Optional[SpanSearchStreamRequestPublicType]

        filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        limit : typing.Optional[int]
            Max number of spans to be streamed

        last_retrieved_id : typing.Optional[str]

        truncate : typing.Optional[bool]
            Truncate image included in either input, output or metadata

        from_time : typing.Optional[dt.datetime]
            Filter spans created from this time (ISO-8601 format).

        to_time : typing.Optional[dt.datetime]
            Filter spans created up to this time (ISO-8601 format). If not provided, defaults to current time. Must be after 'from_time'.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[bytes]
            Spans stream or error during process

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.search_spans()
        asyncio.run(main())
        """
        async with self._raw_client.search_spans(
            trace_id=trace_id,
            project_name=project_name,
            project_id=project_id,
            type=type,
            filters=filters,
            limit=limit,
            last_retrieved_id=last_retrieved_id,
            truncate=truncate,
            from_time=from_time,
            to_time=to_time,
            request_options=request_options,
        ) as r:
            async for data in r.data:
                yield data

    async def update_span_comment(
        self,
        comment_id: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update span comment by id

        Parameters
        ----------
        comment_id : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.spans.update_span_comment(comment_id='commentId', text='text', )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_span_comment(
            comment_id,
            text=text,
            id=id,
            created_at=created_at,
            last_updated_at=last_updated_at,
            created_by=created_by,
            last_updated_by=last_updated_by,
            request_options=request_options,
        )
        return _response.data



import logging
from typing import Any, Optional, cast, Union, Dict

from ..types import BatchFeedbackScoreDict
from ..validation import feedback_score as feedback_score_validator
from .. import logging_messages, llm_usage
from opik.types import LLMProvider


def _is_already_backend_format(usage: Dict[str, Any]) -> bool:
    """Check if usage dict is already in backend-compatible format.

    Backend format has 'original_usage.' prefixed keys for provider-specific data.
    This is used to detect usage data from exports that should be passed through
    without reprocessing.
    """
    return any(key.startswith("original_usage.") for key in usage.keys())


def validate_and_parse_usage(
    usage: Any,
    logger: logging.Logger,
    provider: Optional[Union[LLMProvider, str]],
) -> Optional[Dict[str, int]]:
    if isinstance(usage, llm_usage.OpikUsage):
        return usage.to_backend_compatible_full_usage_dict()

    if usage is None:
        return usage

    # Check if usage is already in backend format (from export/import)
    # If so, return it as-is to preserve the original values
    if isinstance(usage, dict) and _is_already_backend_format(usage):
        # Filter to only keep integer values as expected by backend
        return {k: v for k, v in usage.items() if isinstance(v, int)}

    unknown_provider = (provider is None) or (not LLMProvider.has_value(provider))

    if unknown_provider:
        return llm_usage.build_opik_usage_from_unknown_provider(
            usage
        ).to_backend_compatible_full_usage_dict()

    provider = LLMProvider(provider)

    try:
        opik_usage = llm_usage.build_opik_usage(provider=provider, usage=usage)
        return opik_usage.to_backend_compatible_full_usage_dict()
    except Exception:
        return llm_usage.build_opik_usage_from_unknown_provider(
            usage
        ).to_backend_compatible_full_usage_dict()


def validate_feedback_score(
    feedback_score: Any, logger: logging.Logger
) -> Optional[BatchFeedbackScoreDict]:
    feedback_score_validator_ = feedback_score_validator.FeedbackScoreValidator(
        feedback_score
    )

    if feedback_score_validator_.validate().failed():
        logger.warning(
            logging_messages.INVALID_FEEDBACK_SCORE_WILL_NOT_BE_LOGGED,
            feedback_score,
            feedback_score_validator_.failure_reason_message(),
        )
        return None

    return cast(BatchFeedbackScoreDict, feedback_score)



import logging
from functools import wraps
from typing import Any, Callable, Optional, Tuple, Union, TYPE_CHECKING

from google.adk import models as adk_models

import opik

if TYPE_CHECKING:
    import litellm

LOGGER = logging.getLogger(__name__)


def parse_provider_and_model(
    model: str,
) -> Tuple[Optional[Union[opik.LLMProvider, str]], str]:
    parts = model.split("/", 1)
    if len(parts) != 2:
        return None, parts[0]

    provider = parts[0]
    try:
        provider = opik.LLMProvider(provider)
    except ValueError:
        pass
    return provider, parts[1]


def generate_content_response_decorator(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> adk_models.LlmResponse:
        """
        This wrapper puts token usage data into custom metadata to use it later
        """
        LOGGER.debug("generate_content_response_decorator called")
        result = func(*args, **kwargs)
        model_response = args[0]
        model_response_dict = model_response.to_dict()

        provider_and_model = model_response_dict.get("provider_and_model", None)
        LOGGER.debug(
            "generate_content_response_decorator: provider_and_model=%s",
            provider_and_model,
        )

        if provider_and_model is None:
            return result

        if result.custom_metadata is None:
            LOGGER.debug(
                "generate_content_response_decorator: result.custom_metadata is None, creating new custom metadata"
            )
            result.custom_metadata = {}

        provider, model = parse_provider_and_model(provider_and_model)
        LOGGER.debug(
            "generate_content_response_decorator: provider=%s, model=%s",
            provider,
            model,
        )

        result.custom_metadata["opik_usage"] = model_response_dict["usage"]
        result.custom_metadata["provider"] = provider
        result.custom_metadata["model_version"] = model_response_dict.get(
            "model", model
        )
        LOGGER.debug(
            "generate_content_response_decorator finished: result.custom_metadata=%s",
            result.custom_metadata,
        )
        return result

    return wrapper


def litellm_client_acompletion_decorator(func: Callable) -> Callable:
    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> "litellm.types.utils.ModelResponse":
        """
        this adds more precise provider/model name and it's version
        """
        result = await func(*args, **kwargs)
        result.provider_and_model = kwargs.get("model", None)
        LOGGER.debug(
            "litellm_client_acompletion_decorator called: provider_and_model=%s",
            result.provider_and_model,
        )
        return result

    return wrapper



from abc import abstractmethod
from typing import Any, Dict, Optional, TypeVar

from opik.evaluation.metrics import base_metric

ConfigT = TypeVar("ConfigT")


class BaseSuiteEvaluator(base_metric.BaseMetric):
    """
    Base class for suite evaluators.

    Suite evaluators are designed for use with evaluation suites and provide
    serialization support via `to_config()` and `from_config()` methods for
    compatibility with the backend's online evaluation system.

    Subclasses must implement:
        - `to_config()`: Serialize the evaluator to a configuration object
        - `from_config()`: Create an evaluator instance from a configuration
    """

    @abstractmethod
    def to_config(self) -> Any:
        """
        Serialize the evaluator configuration.

        Returns:
            A configuration object that can be used to reconstruct this evaluator.
        """
        ...

    @classmethod
    @abstractmethod
    def from_config(
        cls,
        config: Any,
        track: bool = True,
        project_name: Optional[str] = None,
        init_kwargs: Optional[Dict[str, Any]] = None,
    ) -> "BaseSuiteEvaluator":
        """
        Create an evaluator instance from a configuration.

        Args:
            config: The configuration object.
            track: Whether to track the evaluator.
            project_name: Optional project name for tracking.
            init_kwargs: Optional dictionary to override __init__ parameters.
                For LLMJudge, this can include 'model' to specify the model name.

        Returns:
            A new evaluator instance.
        """
        ...



import inspect
import logging
from typing import List, Dict, Any, Optional, Callable, Tuple

import opik.exceptions as exceptions
import opik.logging_messages as logging_messages
from opik.api_objects.dataset import dataset_item
from opik.evaluation.metrics import (
    arguments_helpers,
    base_metric,
    score_result,
    arguments_validator,
)
from opik.evaluation.scorers import scorer_wrapper_metric
from opik.evaluation.suite_evaluators import llm_judge
from opik.evaluation.suite_evaluators.llm_judge import config as llm_judge_config
from opik.evaluation.types import ScoringKeyMappingType
from opik.message_processing.emulation import models

from . import exception_analyzer


LOGGER = logging.getLogger(__name__)

EVALUATION_SPAN_PARAMETER_NAME = "task_span"


def _has_evaluation_span_parameter(func: Callable) -> bool:
    """Check if a scoring function expects the task_span parameter."""
    try:
        sig = inspect.signature(func)
        return EVALUATION_SPAN_PARAMETER_NAME in sig.parameters
    except (ValueError, TypeError):
        return False


def split_into_regular_and_task_span_metrics(
    scoring_metrics: List[base_metric.BaseMetric],
) -> Tuple[List[base_metric.BaseMetric], List[base_metric.BaseMetric]]:
    """
    Separate metrics into regular and task-span categories.

    Args:
        scoring_metrics: List of metrics to analyze.

    Returns:
        Tuple of (regular_metrics, task_span_metrics).
    """
    regular_metrics: List[base_metric.BaseMetric] = []
    task_span_metrics: List[base_metric.BaseMetric] = []

    for metric in scoring_metrics:
        if _has_evaluation_span_parameter(metric.score):
            task_span_metrics.append(metric)
        else:
            regular_metrics.append(metric)

    return regular_metrics, task_span_metrics


def _extract_item_evaluators(
    item: dataset_item.DatasetItem,
    evaluator_model: Optional[str],
) -> List[base_metric.BaseMetric]:
    """
    Extract evaluators from dataset item.

    If the item has evaluator configs, instantiate LLMJudge evaluators from them.

    Args:
        item: The dataset item.
        evaluator_model: Optional model name to use for LLMJudge evaluators.

    Returns:
        List of evaluator instances extracted from the item.
    """
    if not item.evaluators:
        return []

    evaluators: List[base_metric.BaseMetric] = []
    for evaluator_item in item.evaluators:
        try:
            if evaluator_item.type == "llm_judge":
                config = llm_judge_config.LLMJudgeConfig(**evaluator_item.config)
                evaluator = llm_judge.LLMJudge.from_config(
                    config, init_kwargs={"model": evaluator_model}
                )
                evaluators.append(evaluator)
            else:
                LOGGER.warning(
                    "Unsupported evaluator type: %s. Only 'llm_judge' is supported.",
                    evaluator_item.type,
                )
        except Exception:
            LOGGER.error(
                "Failed to instantiate evaluator from config: %s",
                evaluator_item.config,
                exc_info=True,
            )
            raise

    return evaluators


def build_metrics_evaluator(
    item: Optional[dataset_item.DatasetItem],
    regular_metrics: List[base_metric.BaseMetric],
    scoring_key_mapping: ScoringKeyMappingType,
    evaluator_model: Optional[str],
) -> "MetricsEvaluator":
    """Build a MetricsEvaluator with suite-level + item-level metrics."""
    all_metrics: List[base_metric.BaseMetric] = list(regular_metrics)
    if item is not None:
        item_evaluators = _extract_item_evaluators(
            item, evaluator_model=evaluator_model
        )
        all_metrics.extend(item_evaluators)

    judges = [m for m in all_metrics if isinstance(m, llm_judge.LLMJudge)]
    non_judges = [m for m in all_metrics if not isinstance(m, llm_judge.LLMJudge)]
    merged = llm_judge.LLMJudge.merged(judges)
    if merged is not None:
        all_metrics = [merged] + non_judges

    return MetricsEvaluator(
        scoring_metrics=all_metrics,
        scoring_key_mapping=scoring_key_mapping,
    )


def _compute_metric_scores(
    scoring_metrics: List[base_metric.BaseMetric],
    mapped_scoring_inputs: Dict[str, Any],
    scoring_key_mapping: ScoringKeyMappingType,
    dataset_item_content: Dict[str, Any],
    task_output: Dict[str, Any],
) -> List[score_result.ScoreResult]:
    """
    Compute scores using given metrics.

    Args:
        scoring_metrics: List of metrics to compute
        mapped_scoring_inputs: Scoring inputs after key mapping (will be used for regular metrics)
        scoring_key_mapping: Mapping for renaming score arguments (empty dict if no mapping)
        dataset_item_content: Dataset item content (will be used for ScorerWrapperMetric)
        task_output: Task output (will be used for ScorerWrapperMetric)

    Returns:
        List of computed score results
    """
    score_results: List[score_result.ScoreResult] = []

    for metric in scoring_metrics:
        try:
            LOGGER.debug("Metric %s score started", metric.name)

            if isinstance(metric, scorer_wrapper_metric.ScorerWrapperMetric):
                # ScorerWrapperMetric uses original dataset item and task output without mappings
                if (
                    task_span := mapped_scoring_inputs.get(
                        EVALUATION_SPAN_PARAMETER_NAME
                    )
                ) is not None:
                    result = metric.score(
                        dataset_item=dataset_item_content,
                        task_outputs=task_output,
                        task_span=task_span,
                    )
                else:
                    result = metric.score(
                        dataset_item=dataset_item_content,
                        task_outputs=task_output,
                    )
            else:
                arguments_validator.validate_score_arguments(
                    metric=metric,
                    kwargs=mapped_scoring_inputs,
                    scoring_key_mapping=scoring_key_mapping,
                )
                result = metric.score(**mapped_scoring_inputs)

            LOGGER.debug("Metric %s score ended", metric.name)

            if isinstance(result, list):
                score_results += result
            else:
                score_results.append(result)

        except exceptions.ScoreMethodMissingArguments:
            raise
        except Exception as exception:
            LOGGER.error(
                "Failed to compute metric %s. Score result will be marked as failed.",
                metric.name,
                exc_info=True,
            )

            if exception_analyzer.is_llm_provider_rate_limit_error(exception):
                LOGGER.error(
                    logging_messages.LLM_PROVIDER_RATE_LIMIT_ERROR_DETECTED_IN_EVALUATE_FUNCTION
                )

            score_results.append(
                score_result.ScoreResult(
                    name=metric.name,
                    value=0.0,
                    reason=str(exception),
                    scoring_failed=True,
                )
            )

    return score_results


class MetricsEvaluator:
    """
    Handles metric computation and scoring.

    Separates metrics into:
    - Regular metrics: Score based on inputs/outputs
    - Task span metrics: Score based on LLM call metadata (tokens, latency, etc)
    """

    def __init__(
        self,
        scoring_metrics: List[base_metric.BaseMetric],
        scoring_key_mapping: ScoringKeyMappingType,
    ):
        self._scoring_key_mapping = scoring_key_mapping
        self._regular_metrics: List[base_metric.BaseMetric] = []
        self._task_span_metrics: List[base_metric.BaseMetric] = []

        self._analyze_metrics(scoring_metrics)

    @property
    def has_task_span_metrics(self) -> bool:
        """Check if any task span scoring metrics are configured."""
        return len(self._task_span_metrics) > 0

    @property
    def task_span_metrics(self) -> List[base_metric.BaseMetric]:
        """Get list of task span scoring metrics."""
        return self._task_span_metrics

    @property
    def regular_metrics(self) -> List[base_metric.BaseMetric]:
        """Get list of regular scoring metrics."""
        return self._regular_metrics

    @property
    def scoring_key_mapping(self) -> ScoringKeyMappingType:
        """Get the scoring key mapping."""
        return self._scoring_key_mapping

    def _analyze_metrics(
        self,
        scoring_metrics: List[base_metric.BaseMetric],
    ) -> None:
        """Separate metrics into regular and task-span categories."""
        self._regular_metrics, self._task_span_metrics = (
            split_into_regular_and_task_span_metrics(scoring_metrics)
        )

        if self.has_task_span_metrics:
            LOGGER.debug(
                "Detected %d LLM task span scoring metrics.",
                len(self._task_span_metrics),
            )

    def compute_regular_scores(
        self,
        dataset_item_content: Dict[str, Any],
        task_output: Dict[str, Any],
    ) -> Tuple[List[score_result.ScoreResult], Dict[str, Any]]:
        """
        Compute scores using regular metrics.

        Args:
            dataset_item_content: Dataset item content
            task_output: Task output

        Returns:
            Tuple of (score results, mapped scoring inputs used for scoring regular non-wrapper metrics)
        """
        mapped_scoring_inputs = arguments_helpers.create_scoring_inputs(
            dataset_item=dataset_item_content,
            task_output=task_output,
            scoring_key_mapping=self._scoring_key_mapping,
        )

        score_results = _compute_metric_scores(
            scoring_metrics=self._regular_metrics,
            mapped_scoring_inputs=mapped_scoring_inputs,
            scoring_key_mapping=self._scoring_key_mapping,
            dataset_item_content=dataset_item_content,
            task_output=task_output,
        )

        return score_results, mapped_scoring_inputs

    def compute_task_span_scores(
        self,
        dataset_item_content: Dict[str, Any],
        task_output: Dict[str, Any],
        task_span: models.SpanModel,
    ) -> Tuple[List[score_result.ScoreResult], Dict[str, Any]]:
        """
        Compute scores using task span metrics.

        Args:
            dataset_item_content: Dataset item content
            task_output: Task output
            task_span: Span model containing task execution metadata

        Returns:
            Tuple of (score results, mapped scoring inputs used for scoring regular non-wrapper metrics)
        """
        mapped_scoring_inputs = arguments_helpers.create_scoring_inputs(
            dataset_item=dataset_item_content,
            task_output=task_output,
            scoring_key_mapping=self._scoring_key_mapping,
        )

        mapped_scoring_inputs_with_span = {
            **mapped_scoring_inputs,
            EVALUATION_SPAN_PARAMETER_NAME: task_span,
        }

        score_results = _compute_metric_scores(
            scoring_metrics=self._task_span_metrics,
            mapped_scoring_inputs=mapped_scoring_inputs_with_span,
            scoring_key_mapping=self._scoring_key_mapping,
            dataset_item_content=dataset_item_content,
            task_output=task_output,
        )

        return score_results, mapped_scoring_inputs_with_span



import json
from typing import List, Optional

from . import experiment_item
from .. import rest_stream_parser
from ... import exceptions, rest_api
from ...rest_api.types import experiment_public


def get_experiment_data_by_name(
    rest_client: rest_api.OpikApi,
    name: str,
    project_name: Optional[str],
) -> experiment_public.ExperimentPublic:
    # TODO: this method is deprecated and should be removed after
    #  deprecated Opik.get_experiment_by_name() will be removed.
    #  This function should not be used anywhere else except for deprecated logic as it is confusing and misleading

    experiments = get_experiments_data_by_name(
        rest_client, name, project_name=project_name
    )
    for experiment in experiments:
        if experiment.name == name:
            return experiment

    raise exceptions.ExperimentNotFound(f"No experiment found with the name '{name}'.")


def get_experiments_data_by_name(
    rest_client: rest_api.OpikApi,
    name: str,
    project_name: Optional[str],
    max_results: Optional[int] = None,
) -> List[experiment_public.ExperimentPublic]:
    experiments = rest_stream_parser.read_and_parse_full_stream(
        read_source=lambda current_batch_size,
        last_retrieved_id: rest_client.experiments.stream_experiments(
            name=name,
            limit=current_batch_size,
            last_retrieved_id=last_retrieved_id,
            project_name=project_name,
        ),
        max_results=max_results,
        parsed_item_class=experiment_public.ExperimentPublic,
    )

    if len(experiments) == 0:
        raise exceptions.ExperimentNotFound(
            f"No experiment(s) found with the name '{name}'."
        )

    return experiments


def find_experiment_items_for_dataset(
    rest_client: rest_api.OpikApi,
    dataset_id: str,
    experiment_ids: List[str],
    max_results: int,
    truncate: bool,
    filter_expression: Optional[str] = None,
) -> List[experiment_item.ExperimentItemContent]:
    PAGE_SIZE = 100

    collected_items: List[experiment_item.ExperimentItemContent] = []
    experiment_ids_json = json.dumps(experiment_ids)

    page_number = 1
    while len(collected_items) < max_results:
        page_dataset_items = (
            rest_client.datasets.find_dataset_items_with_experiment_items(
                id=dataset_id,
                page=page_number,
                size=PAGE_SIZE,
                experiment_ids=experiment_ids_json,
                truncate=truncate,
                filters=filter_expression,
            )
        )

        if not page_dataset_items.content:
            break

        # Build a flat list of experiment-item compares for this page in a readable way
        experiment_items = []
        for dataset_item in page_dataset_items.content:
            # Guard if dataset_item.experiment_items might be None
            if dataset_item.experiment_items is not None:
                for experiment_item_compare in dataset_item.experiment_items:
                    # Convert to domain objects
                    dataset_item_data = dataset_item.data
                    if dataset_item_data is not None:
                        dataset_item_data.update({"id": dataset_item.id})
                    experiment_item_content = experiment_item.ExperimentItemContent.from_rest_experiment_item_compare(
                        value=experiment_item_compare,
                        dataset_item_data=dataset_item_data,
                    )

                    experiment_items.append(experiment_item_content)

        if not experiment_items:
            page_number += 1
            continue

        remaining = max_results - len(collected_items)

        collected_items.extend(experiment_items[:remaining])

        page_number += 1

    return collected_items



from typing import List, Optional

import dataclasses

from . import test_case
from .metrics import score_result


@dataclasses.dataclass
class TestResult:
    test_case: test_case.TestCase
    score_results: List[score_result.ScoreResult]
    trial_id: int
    task_execution_time: Optional[float] = None
    scoring_time: Optional[float] = None



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.retention_rule_page_public import RetentionRulePagePublic
from ..types.retention_rule_public import RetentionRulePublic
from .raw_client import AsyncRawRetentionRulesClient, RawRetentionRulesClient
from .types.retention_rule_write_retention import RetentionRuleWriteRetention

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RetentionRulesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawRetentionRulesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawRetentionRulesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawRetentionRulesClient
        """
        return self._raw_client

    def find_retention_rules(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        include_inactive: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> RetentionRulePagePublic:
        """
        List retention rules for the caller's workspace. Defaults to active only.

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        include_inactive : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RetentionRulePagePublic
            Retention rules page

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.retention_rules.find_retention_rules()
        """
        _response = self._raw_client.find_retention_rules(
            page=page, size=size, include_inactive=include_inactive, request_options=request_options
        )
        return _response.data

    def create_retention_rule(
        self,
        *,
        retention: RetentionRuleWriteRetention,
        project_id: typing.Optional[str] = OMIT,
        organization_level: typing.Optional[bool] = OMIT,
        apply_to_past: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> RetentionRulePublic:
        """
        Create a new retention rule. Auto-deactivates any existing active rule for the same scope.

        Parameters
        ----------
        retention : RetentionRuleWriteRetention

        project_id : typing.Optional[str]

        organization_level : typing.Optional[bool]
            Set to true to create an organization-level rule

        apply_to_past : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RetentionRulePublic
            Created

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.retention_rules.create_retention_rule(retention="short_14d", )
        """
        _response = self._raw_client.create_retention_rule(
            retention=retention,
            project_id=project_id,
            organization_level=organization_level,
            apply_to_past=apply_to_past,
            request_options=request_options,
        )
        return _response.data

    def get_retention_rule_by_id(
        self, rule_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> RetentionRulePublic:
        """
        Get a specific retention rule by id

        Parameters
        ----------
        rule_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RetentionRulePublic
            Retention rule

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.retention_rules.get_retention_rule_by_id(rule_id='ruleId', )
        """
        _response = self._raw_client.get_retention_rule_by_id(rule_id, request_options=request_options)
        return _response.data

    def deactivate_retention_rule(
        self, rule_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Soft-deactivate a retention rule (sets enabled=false). Rules are never hard-deleted for audit trail.

        Parameters
        ----------
        rule_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.retention_rules.deactivate_retention_rule(rule_id='ruleId', )
        """
        _response = self._raw_client.deactivate_retention_rule(rule_id, request_options=request_options)
        return _response.data


class AsyncRetentionRulesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawRetentionRulesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawRetentionRulesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawRetentionRulesClient
        """
        return self._raw_client

    async def find_retention_rules(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        include_inactive: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> RetentionRulePagePublic:
        """
        List retention rules for the caller's workspace. Defaults to active only.

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        include_inactive : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RetentionRulePagePublic
            Retention rules page

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.retention_rules.find_retention_rules()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_retention_rules(
            page=page, size=size, include_inactive=include_inactive, request_options=request_options
        )
        return _response.data

    async def create_retention_rule(
        self,
        *,
        retention: RetentionRuleWriteRetention,
        project_id: typing.Optional[str] = OMIT,
        organization_level: typing.Optional[bool] = OMIT,
        apply_to_past: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> RetentionRulePublic:
        """
        Create a new retention rule. Auto-deactivates any existing active rule for the same scope.

        Parameters
        ----------
        retention : RetentionRuleWriteRetention

        project_id : typing.Optional[str]

        organization_level : typing.Optional[bool]
            Set to true to create an organization-level rule

        apply_to_past : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RetentionRulePublic
            Created

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.retention_rules.create_retention_rule(retention="short_14d", )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_retention_rule(
            retention=retention,
            project_id=project_id,
            organization_level=organization_level,
            apply_to_past=apply_to_past,
            request_options=request_options,
        )
        return _response.data

    async def get_retention_rule_by_id(
        self, rule_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> RetentionRulePublic:
        """
        Get a specific retention rule by id

        Parameters
        ----------
        rule_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        RetentionRulePublic
            Retention rule

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.retention_rules.get_retention_rule_by_id(rule_id='ruleId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_retention_rule_by_id(rule_id, request_options=request_options)
        return _response.data

    async def deactivate_retention_rule(
        self, rule_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Soft-deactivate a retention rule (sets enabled=false). Rules are never hard-deleted for audit trail.

        Parameters
        ----------
        rule_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.retention_rules.deactivate_retention_rule(rule_id='ruleId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.deactivate_retention_rule(rule_id, request_options=request_options)
        return _response.data



"""Prompt import functionality."""

import json
from pathlib import Path
from typing import Dict, Optional

import opik
from opik.api_objects.prompt import Prompt, ChatPrompt
from opik.api_objects.prompt.types import PromptType
from rich.console import Console

from ..migration_manifest import MigrationManifest
from .utils import matches_name_pattern

console = Console()


def import_prompts_from_directory(
    client: opik.Opik,
    source_dir: Path,
    dry_run: bool,
    name_pattern: Optional[str],
    debug: bool,
    manifest: Optional[MigrationManifest] = None,
) -> Dict[str, int]:
    """Import prompts from a directory.

    Returns:
        Dictionary with keys: 'prompts', 'prompts_skipped', 'prompts_errors'
    """
    try:
        prompt_files = list(source_dir.glob("prompt_*.json"))

        if not prompt_files:
            console.print("[yellow]No prompt files found in the directory[/yellow]")
            return {"prompts": 0, "prompts_skipped": 0, "prompts_errors": 0}

        imported_count = 0
        skipped_count = 0
        error_count = 0
        for prompt_file in prompt_files:
            try:
                # Skip files already completed in a previous run
                if manifest and not dry_run and manifest.is_file_completed(prompt_file):
                    if debug:
                        console.print(
                            f"[blue]Skipping {prompt_file.name} (already imported in a previous run)[/blue]"
                        )
                    skipped_count += 1
                    continue

                with open(prompt_file, "r", encoding="utf-8") as f:
                    prompt_data = json.load(f)

                # Handle two export formats:
                # 1. {"name": "...", "current_version": {...}, "history": [...]} - from export_single_prompt
                # 2. {"prompt": {"name": "...", ...}, "current_version": {...}, "history": [...]} - from export_experiment_prompts
                prompt_name = prompt_data.get("name") or (
                    prompt_data.get("prompt", {}).get("name")
                    if prompt_data.get("prompt")
                    else None
                )

                # Check if name is missing or empty
                if not prompt_name or (
                    isinstance(prompt_name, str) and not prompt_name.strip()
                ):
                    console.print(
                        f"[yellow]Skipping {prompt_file.name} (no name found)[/yellow]"
                    )
                    skipped_count += 1
                    continue

                # Strip whitespace from name
                prompt_name = prompt_name.strip()

                # Filter by name pattern if specified
                if name_pattern and not matches_name_pattern(prompt_name, name_pattern):
                    if debug:
                        console.print(
                            f"[blue]Skipping prompt {prompt_name} (doesn't match pattern)[/blue]"
                        )
                    skipped_count += 1
                    continue

                if dry_run:
                    console.print(f"[blue]Would import prompt: {prompt_name}[/blue]")
                    imported_count += 1
                    continue

                if debug:
                    console.print(f"[blue]Importing prompt: {prompt_name}[/blue]")

                # Get current version data
                current_version = prompt_data.get("current_version", {})
                prompt_content = current_version.get("prompt")
                metadata = current_version.get("metadata")
                prompt_type = current_version.get("type")
                template_structure = current_version.get("template_structure", "text")

                # Validate prompt content exists
                if prompt_content is None:
                    console.print(
                        f"[yellow]Skipping {prompt_name} (no prompt content found)[/yellow]"
                    )
                    skipped_count += 1
                    continue

                # Convert string type to PromptType enum if needed
                if prompt_type and isinstance(prompt_type, str):
                    try:
                        prompt_type_enum = PromptType(prompt_type)
                    except ValueError:
                        console.print(
                            f"[yellow]Unknown prompt type '{prompt_type}', using MUSTACHE[/yellow]"
                        )
                        prompt_type_enum = PromptType.MUSTACHE
                else:
                    prompt_type_enum = PromptType.MUSTACHE

                # Create the prompt based on template_structure
                try:
                    if template_structure == "chat":
                        # ChatPrompt expects a list of message dictionaries
                        if not isinstance(prompt_content, list):
                            console.print(
                                f"[yellow]Skipping {prompt_name} (chat prompt content must be a list of messages)[/yellow]"
                            )
                            skipped_count += 1
                            continue

                        # Create ChatPrompt
                        ChatPrompt(
                            name=prompt_name,
                            messages=prompt_content,
                            metadata=metadata,
                            type=prompt_type_enum,
                        )
                        if debug:
                            console.print(
                                f"[green]Imported chat prompt: {prompt_name}[/green]"
                            )
                    else:
                        # Text Prompt expects a string
                        if not isinstance(prompt_content, str):
                            console.print(
                                f"[yellow]Skipping {prompt_name} (text prompt content must be a string)[/yellow]"
                            )
                            skipped_count += 1
                            continue

                        # Create Prompt
                        Prompt(
                            name=prompt_name,
                            prompt=prompt_content,
                            metadata=metadata,
                            type=prompt_type_enum,
                        )
                        if debug:
                            console.print(
                                f"[green]Imported text prompt: {prompt_name}[/green]"
                            )

                    imported_count += 1

                    # Mark this file as completed in the manifest
                    if manifest:
                        manifest.mark_file_completed(prompt_file)

                except Exception as e:
                    console.print(
                        f"[red]Error creating prompt {prompt_name}: {e}[/red]"
                    )
                    if manifest:
                        manifest.mark_file_failed(prompt_file, str(e))
                    error_count += 1
                    continue

            except Exception as e:
                console.print(
                    f"[red]Error importing prompt from {prompt_file}: {e}[/red]"
                )
                if manifest:
                    manifest.mark_file_failed(prompt_file, str(e))
                error_count += 1
                continue

        return {
            "prompts": imported_count,
            "prompts_skipped": skipped_count,
            "prompts_errors": error_count,
        }

    except Exception as e:
        console.print(f"[red]Error importing prompts: {e}[/red]")
        return {"prompts": 0, "prompts_skipped": 0, "prompts_errors": 1}



from typing import Union, List, Any

from . import types
from .. import base_metric, score_result


class ConversationThreadMetric(base_metric.BaseMetric):
    """
    Abstract base class for all conversation thread metrics. When creating a custom
    conversation metric, you should inherit from this class and implement the abstract methods.

    Conversation metrics are designed to evaluate multi-turn conversations rather than
    single input-output pairs. They accept a conversation as a list of message dictionaries,
    where each message has a 'role' (either 'user' or 'assistant') and 'content'.

    Args:
        name: The name of the metric. If not provided, uses the class name as default.
        track: Whether to track the metric. Defaults to True.
        project_name: Optional project name to track the metric in for the cases when
            there is no parent span/trace to inherit project name from.

    Example:
        >>> from opik.evaluation.metrics.conversation import conversation_thread_metric, types
        >>> from opik.evaluation.metrics import score_result
        >>> from typing import Any
        >>>
        >>> class ConversationLengthMetric(conversation_thread_metric.ConversationThreadMetric):
        >>>     def __init__(self, name: str = "conversation_length_score"):
        >>>         super().__init__(name)
        >>>
        >>>     def score(self, conversation: types.Conversation, **kwargs: Any):
        >>>         num_turns = sum(1 for msg in conversation if msg["role"] == "assistant")
        >>>         return score_result.ScoreResult(
        >>>             name=self.name,
        >>>             value=num_turns,
        >>>             reason=f"Conversation has {num_turns} turns"
        >>>         )
    """

    def score(
        self, conversation: types.Conversation, **kwargs: Any
    ) -> Union[score_result.ScoreResult, List[score_result.ScoreResult]]:
        """
        Evaluate a conversation and return a score.

        Args:
            conversation: A list of conversation messages. Each message is a dictionary
                with 'role' (either 'user' or 'assistant') and 'content' (the message text).
            **kwargs: Additional keyword arguments that may be used by specific metric implementations.

        Returns:
            A ScoreResult object or list of ScoreResult objects containing the evaluation score,
            metric name, and optional reasoning.
        """
        raise NotImplementedError(
            "Please use concrete metric classes instead of this one."
        )

    async def ascore(
        self, conversation: types.Conversation, **kwargs: Any
    ) -> Union[score_result.ScoreResult, List[score_result.ScoreResult]]:
        """
        Asynchronously evaluate a conversation and return a score.

        This is the async version of the score method. By default, it calls the
        synchronous score method, but can be overridden for true async implementations.

        Args:
            conversation: A list of conversation messages. Each message is a dictionary
                with 'role' (either 'user' or 'assistant') and 'content' (the message text).
            **kwargs: Additional keyword arguments that may be used by specific metric implementations.

        Returns:
            A ScoreResult object or list of ScoreResult objects containing the evaluation score,
            metric name, and optional reasoning.
        """
        raise NotImplementedError(
            "Please use concrete metric classes instead of this one."
        )



from typing import Optional, Any, Dict

from rich.console import Console

from opik import config
from opik.cli.healthcheck import opik_rest_helpers, rich_representation

console = Console()


def run(api_key: Optional[str], workspace: Optional[str]) -> None:
    rich_representation.print_header("user permissions")

    settings = config.OpikConfig()
    api_key = api_key or settings.api_key
    workspace = workspace or settings.workspace

    if api_key is None:
        raise ValueError("API key is required")
    if workspace is None:
        raise ValueError("Workspace is required")

    try:
        permissions = get_user_permissions(
            api_key, workspace=workspace, url=settings.url_override
        )
        rich_representation.print_user_permissions(permissions)
    except ConnectionError as e:
        err_msg = f"Failed to fetch user permissions:\n{e}"
        rich_representation.print_error_message(err_msg)


def get_user_permissions(api_key: str, workspace: str, url: str) -> Dict[str, Any]:
    rich_representation.print_opik_permissions_url(opik_url=url)
    return opik_rest_helpers.list_user_permissions(
        api_key, workspace=workspace, url=url
    )



import atexit
import logging
import os
import threading
import time
import typing

from .blueprint import Blueprint

logger = logging.getLogger(__name__)

DEFAULT_TTL_SECONDS = 300
_MIN_REFRESH_INTERVAL_SECONDS = 1.0

_CacheKey = typing.Tuple[str, typing.Optional[str], typing.Optional[str]]


def _get_ttl_seconds() -> int:
    raw = os.environ.get("OPIK_CONFIG_TTL_SECONDS")
    if raw is not None:
        try:
            return int(raw)
        except ValueError:
            pass
    return DEFAULT_TTL_SECONDS


class SharedConfigCache:
    def __init__(self, ttl_seconds: int = DEFAULT_TTL_SECONDS) -> None:
        self._lock = threading.RLock()
        self.blueprint_id: typing.Optional[str] = None
        self.blueprint_version: typing.Optional[str] = None
        self.values: typing.Dict[str, typing.Any] = {}
        self._registered_field_types: typing.Dict[str, typing.Any] = {}
        self._ttl_seconds = ttl_seconds
        self._last_fetch: typing.Optional[float] = None
        self._refresh_callback: typing.Optional[
            typing.Callable[[], typing.Optional[Blueprint]]
        ] = None

    def set_refresh_callback(
        self, callback: typing.Callable[[], typing.Optional[Blueprint]]
    ) -> None:
        with self._lock:
            if self._refresh_callback is None:
                self._refresh_callback = callback

    def register_fields(
        self, prefixed_field_types: typing.Dict[str, typing.Any]
    ) -> None:
        with self._lock:
            self._registered_field_types.update(prefixed_field_types)

    @property
    def all_field_types(self) -> typing.Dict[str, typing.Any]:
        with self._lock:
            return dict(self._registered_field_types)

    def update(self, blueprint: Blueprint) -> None:
        new_values = dict(blueprint._values)
        with self._lock:
            self.blueprint_id = blueprint.id
            self.blueprint_version = blueprint.name
            self.values = new_values
            self._last_fetch = time.monotonic()

    def value_keys(self) -> typing.Set[str]:
        with self._lock:
            return set(self.values.keys())

    def is_stale(self) -> bool:
        with self._lock:
            if self._last_fetch is None:
                return True
            return (time.monotonic() - self._last_fetch) >= self._ttl_seconds

    def try_background_refresh(self) -> None:
        with self._lock:
            callback = self._refresh_callback
        if callback is None:
            return
        try:
            bp = callback()
            if bp is not None:
                self.update(bp)
        except Exception:
            logger.debug("Background cache refresh failed", exc_info=True)


class CacheRefreshThread(threading.Thread):
    def __init__(
        self,
        get_caches: typing.Callable[[], typing.List[SharedConfigCache]],
        interval_seconds: typing.Optional[float] = None,
    ) -> None:
        super().__init__(daemon=True, name="OpikCacheRefresh")
        self._get_caches = get_caches
        self._stop_event = threading.Event()
        self._interval = interval_seconds

    def run(self) -> None:
        while not self._stop_event.is_set():
            self._refresh_all_stale()
            interval = self._interval or float(_get_ttl_seconds())
            self._stop_event.wait(max(interval, _MIN_REFRESH_INTERVAL_SECONDS))

    def _refresh_all_stale(self) -> None:
        for cache in self._get_caches():
            if self._stop_event.is_set():
                break
            if cache.is_stale():
                cache.try_background_refresh()

    def close(self) -> None:
        self._stop_event.set()


class SharedCacheRegistry:
    def __init__(self) -> None:
        self._lock = threading.RLock()
        self._caches: typing.Dict[_CacheKey, SharedConfigCache] = {}
        self._thread: typing.Optional[CacheRefreshThread] = None
        self._thread_lock = threading.Lock()

    def get(
        self,
        project_name: str,
        env: typing.Optional[str],
        mask_id: typing.Optional[str],
    ) -> SharedConfigCache:
        key: _CacheKey = (project_name, env, mask_id)
        with self._lock:
            if key not in self._caches:
                self._caches[key] = SharedConfigCache(ttl_seconds=_get_ttl_seconds())
            return self._caches[key]

    def clear(self) -> None:
        self.stop_refresh_thread()
        with self._lock:
            self._caches.clear()

    def ensure_refresh_thread_started(self) -> None:
        with self._thread_lock:
            if self._thread is not None and self._thread.is_alive():
                return
            self._thread = CacheRefreshThread(
                get_caches=lambda: list(self._caches.values())
            )
            self._thread.start()
            atexit.register(self.stop_refresh_thread)

    def stop_refresh_thread(self) -> None:
        with self._thread_lock:
            if self._thread is not None:
                thread = self._thread
                thread.close()
                try:
                    thread.join(timeout=5)
                    if thread.is_alive():
                        logger.error(
                            "Cache refresh thread did not stop within the timeout."
                        )
                except Exception:
                    logger.exception(
                        "Unexpected error while waiting for cache refresh thread to stop."
                    )
                self._thread = None


_registry = SharedCacheRegistry()


def get_global_registry() -> SharedCacheRegistry:
    return _registry


# ---------------------------------------------------------------------------
# Module-level helpers (used by base.py and opik_client.py)
# ---------------------------------------------------------------------------


def get_cached_config(
    project_name: str,
    env: typing.Optional[str],
    mask_id: typing.Optional[str],
) -> SharedConfigCache:
    return _registry.get(project_name, env, mask_id)


def init_cache_entry(
    project_name: str,
    env: typing.Optional[str],
    mask_id: typing.Optional[str],
    prefixed_field_types: typing.Dict[str, typing.Any],
    agent_config_manager: typing.Any,
    blueprint: typing.Optional[Blueprint] = None,
) -> None:
    shared_cache = _registry.get(project_name, env, mask_id)
    shared_cache.register_fields(prefixed_field_types)

    if blueprint is not None:
        shared_cache.update(blueprint)

    if agent_config_manager is not None and mask_id is None:

        def _refresh() -> typing.Optional[Blueprint]:
            return agent_config_manager.get_blueprint(
                env=env,
                mask_id=mask_id,
                field_types=shared_cache.all_field_types,
            )

        shared_cache.set_refresh_callback(_refresh)
        _registry.ensure_refresh_thread_started()



# This file was auto-generated by Fern from our API Definition.

import contextlib
import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.bad_request_error import BadRequestError
from ..errors.not_found_error import NotFoundError
from ..errors.too_many_requests_error import TooManyRequestsError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.error_message import ErrorMessage

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawOllieStateClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    @contextlib.contextmanager
    def download_ollie_state(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Iterator[HttpResponse[typing.Iterator[bytes]]]:
        """
        Download stored ollie state file

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[HttpResponse[typing.Iterator[bytes]]]
            Ollie state file
        """
        with self._client_wrapper.httpx_client.stream(
            "v1/private/ollie/state",
            method="GET",
            request_options=request_options,
        ) as _response:

            def stream() -> HttpResponse[typing.Iterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return HttpResponse(
                            response=_response, data=(_chunk for _chunk in _response.iter_bytes(chunk_size=_chunk_size))
                        )
                    _response.read()
                    if _response.status_code == 401:
                        raise UnauthorizedError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 404:
                        raise NotFoundError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield stream()

    def replace_ollie_state(
        self,
        *,
        request: typing.Union[bytes, typing.Iterator[bytes], typing.AsyncIterator[bytes]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Upload gzip-compressed SQLite DB file, replacing any existing state

        Parameters
        ----------
        request : typing.Union[bytes, typing.Iterator[bytes], typing.AsyncIterator[bytes]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/ollie/state",
            method="PUT",
            content=request,
            headers={
                "content-type": "application/gzip",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_ollie_state(self, *, request_options: typing.Optional[RequestOptions] = None) -> HttpResponse[None]:
        """
        Clear stored ollie state

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/ollie/state",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawOllieStateClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    @contextlib.asynccontextmanager
    async def download_ollie_state(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]:
        """
        Download stored ollie state file

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]
            Ollie state file
        """
        async with self._client_wrapper.httpx_client.stream(
            "v1/private/ollie/state",
            method="GET",
            request_options=request_options,
        ) as _response:

            async def stream() -> AsyncHttpResponse[typing.AsyncIterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return AsyncHttpResponse(
                            response=_response,
                            data=(_chunk async for _chunk in _response.aiter_bytes(chunk_size=_chunk_size)),
                        )
                    await _response.aread()
                    if _response.status_code == 401:
                        raise UnauthorizedError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 404:
                        raise NotFoundError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield await stream()

    async def replace_ollie_state(
        self,
        *,
        request: typing.Union[bytes, typing.Iterator[bytes], typing.AsyncIterator[bytes]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Upload gzip-compressed SQLite DB file, replacing any existing state

        Parameters
        ----------
        request : typing.Union[bytes, typing.Iterator[bytes], typing.AsyncIterator[bytes]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/ollie/state",
            method="PUT",
            content=request,
            headers={
                "content-type": "application/gzip",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_ollie_state(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Clear stored ollie state

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/ollie/state",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



from typing import List, Callable, Optional

from . import evaluation_result
from ...api_objects import opik_client
from ...api_objects.conversation import conversation_thread, conversation_factory
from ...rest_api import TraceThread, JsonListStringPublic
from ...types import BatchFeedbackScoreDict
from ...api_objects.threads import threads_client


def log_feedback_scores(
    results: List[evaluation_result.ThreadEvaluationResult],
    project_name: Optional[str],
    client: threads_client.ThreadsClient,
) -> None:
    for result in results:
        feedback_scores = [
            BatchFeedbackScoreDict(
                id=result.thread_id,
                name=score.name,
                value=score.value,
                reason=score.reason,
            )
            for score in result.scores
            if not score.scoring_failed
        ]
        client.log_threads_feedback_scores(
            scores=feedback_scores,
            project_name=project_name,
        )


def load_conversation_thread(
    thread: TraceThread,
    trace_input_transform: Callable[[JsonListStringPublic], str],
    trace_output_transform: Callable[[JsonListStringPublic], str],
    max_results: int,
    project_name: Optional[str],
    client: opik_client.Opik,
) -> conversation_thread.ConversationThread:
    traces = client.search_traces(
        project_name=project_name,
        filter_string=f'thread_id = "{thread.id}"',
        max_results=max_results,
        truncate=False,
    )
    return conversation_factory.create_conversation_from_traces(
        traces=traces,
        input_transform=trace_input_transform,
        output_transform=trace_output_transform,
    )



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .feedback_score_experiment_item_bulk_write_view import FeedbackScoreExperimentItemBulkWriteView
from .json_list_string_experiment_item_bulk_write_view import JsonListStringExperimentItemBulkWriteView
from .span_experiment_item_bulk_write_view import SpanExperimentItemBulkWriteView
from .trace_experiment_item_bulk_write_view import TraceExperimentItemBulkWriteView


class ExperimentItemBulkRecordExperimentItemBulkWriteView(UniversalBaseModel):
    dataset_item_id: str
    evaluate_task_result: typing.Optional[JsonListStringExperimentItemBulkWriteView] = None
    trace: typing.Optional[TraceExperimentItemBulkWriteView] = None
    spans: typing.Optional[typing.List[SpanExperimentItemBulkWriteView]] = None
    feedback_scores: typing.Optional[typing.List[FeedbackScoreExperimentItemBulkWriteView]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""
LLMJudge evaluator for evaluation suites.

This module provides an LLM-as-a-judge evaluator that can be stored
in the backend and used with evaluation suites. The evaluator can
evaluate one or more assertions/criteria against the agent's output.
"""

import logging
from typing import Any, Dict, List, Optional

import tenacity

from opik.evaluation.models import models_factory
from opik.evaluation.metrics import score_result
from opik.exceptions import LLMJudgeParseError

from opik.evaluation.suite_evaluators import base
from . import config as llm_judge_config
from . import parsers

LOGGER = logging.getLogger(__name__)

_RETRY_POLICY = tenacity.retry(
    retry=tenacity.retry_if_exception_type(LLMJudgeParseError),
    stop=tenacity.stop_after_attempt(3),
    before_sleep=tenacity.before_sleep_log(LOGGER, logging.WARNING),
    reraise=True,
)


LLM_JUDGE_SYSTEM_PROMPT = """You are an expert judge tasked with evaluating if an AI agent's output satisfies a set of assertions.

For each assertion, provide:
- score: true if the assertion passes, false if it fails
- reason: A brief explanation of your judgment
- confidence: A float between 0.0 and 1.0 indicating how confident you are in your judgment
"""

LLM_JUDGE_USER_TEMPLATE = """## Input
The INPUT section contains all data that the agent received. This may include the actual user query, conversation history, context, metadata, or other structured information. Identify the core user request within this data.

---BEGIN INPUT---
{input}
---END INPUT---

## Output
The OUTPUT section contains all data produced by the agent. This may include the agent's response text, tool calls, intermediate results, metadata, or other structured information. Focus on the substantive response when evaluating assertions.

---BEGIN OUTPUT---
{output}
---END OUTPUT---

## Assertions
Evaluate each of the following assertions against the agent's output.
Use the provided field key as the JSON property name for each assertion result.

{assertions}
"""


def _format_value(value: Any) -> str:
    """Format a value for inclusion in the prompt."""
    if isinstance(value, str):
        return value
    import json

    return json.dumps(value, indent=2, default=str)


def _generate_prompt(
    input: Any,
    output: Any,
    assertions_text: str,
) -> str:
    """Generate the LLM query for evaluating assertions.

    Combines the system prompt and user template into a single string
    because the model's generate_string API accepts only one input string.
    """
    user_content = LLM_JUDGE_USER_TEMPLATE.format(
        input=_format_value(input),
        output=_format_value(output),
        assertions=assertions_text,
    )
    return LLM_JUDGE_SYSTEM_PROMPT + "\n" + user_content


class LLMJudge(base.BaseSuiteEvaluator):
    """
    LLM-as-a-judge evaluator for evaluation suites.

    This evaluator uses an LLM to judge whether an agent's output satisfies
    one or more assertions/criteria. It returns a ScoreResult for each assertion.

    The evaluator configuration can be serialized via `to_config()` and
    `from_config()` for storage in the backend's online evaluation system.

    Args:
        assertions: A list of assertion strings describing expected behaviors.
            Each string should describe what the output should satisfy.
        name: The name of the evaluator (used as prefix for score names).
            Defaults to "llm_judge".
        model: The model name to use for evaluation. If not provided,
            uses the default model (gpt-5-nano).
        track: Whether to track the evaluator. Defaults to True.
        project_name: Optional project name for tracking.
        seed: Optional seed value for reproducible model generation.
        temperature: Optional temperature value for model generation.
        reasoning_effort: Optional reasoning effort level for the model.
            Supported values depend on the provider (e.g., "low", "medium", "high"
            for OpenAI reasoning models).

    Example:
        >>> from opik.evaluation.suite_evaluators import LLMJudge
        >>>
        >>> evaluator = LLMJudge(
        ...     assertions=[
        ...         "Response does not contain hallucinated information",
        ...         "Response is helpful to the user",
        ...     ]
        ... )
        >>>
        >>> # Score an input/output pair
        >>> results = evaluator.score(
        ...     input="What is the capital of France?",
        ...     output="The capital of France is Paris."
        ... )
        >>> for result in results:
        ...     print(f"{result.name}: {result.value}")
    """

    def __init__(
        self,
        assertions: List[str],
        name: str = "llm_judge",
        model: Optional[str] = None,
        track: bool = True,
        project_name: Optional[str] = None,
        seed: Optional[int] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[str] = None,
    ):
        super().__init__(name=name, track=track, project_name=project_name)

        self._assertions: List[str] = list(assertions)

        self._seed = seed
        self._temperature = temperature
        self._reasoning_effort = reasoning_effort
        self._model_name: str = model or llm_judge_config.DEFAULT_MODEL_NAME
        self._reasoning_effort = (
            reasoning_effort or llm_judge_config.DEFAULT_REASONING_EFFORT
        )
        self._init_model()

    @property
    def assertions(self) -> List[str]:
        """The assertions this evaluator checks (read-only copy)."""
        return list(self._assertions)

    def _has_same_settings(self, other: "LLMJudge") -> bool:
        return (
            self._model_name == other._model_name
            and self._temperature == other._temperature
            and self._seed == other._seed
            and self._reasoning_effort == other._reasoning_effort
            and self.track == other.track
        )

    @classmethod
    def merged(cls, judges: List["LLMJudge"]) -> Optional["LLMJudge"]:
        """Create a single LLMJudge by merging assertions from multiple judges.

        Uses settings (model, temperature, seed, track) from the first judge.
        Duplicate assertions are removed while preserving order.

        Returns None if fewer than 2 judges are provided or if their settings
        (model, temperature, seed, track) differ.
        """
        if len(judges) <= 1:
            return None

        first = judges[0]
        if not all(first._has_same_settings(j) for j in judges[1:]):
            return None

        seen: set = set()
        merged_assertions: List[str] = []
        for judge in judges:
            for assertion in judge._assertions:
                if assertion not in seen:
                    seen.add(assertion)
                    merged_assertions.append(assertion)

        return cls(
            assertions=merged_assertions,
            name=first.name,
            model=first._model_name,
            track=first.track,
            seed=first._seed,
            temperature=first._temperature,
            reasoning_effort=first._reasoning_effort,
        )

    def _init_model(self) -> None:
        model_kwargs: Dict[str, Any] = {}
        if self._temperature is not None:
            model_kwargs["temperature"] = self._temperature
        if self._seed is not None:
            model_kwargs["seed"] = self._seed
        if self._reasoning_effort is not None:
            model_kwargs["reasoning_effort"] = self._reasoning_effort

        self._model = models_factory.get(
            model_name=self._model_name, track=self.track, **model_kwargs
        )

    def score(
        self,
        input: Any,
        output: Any,
        **ignored_kwargs: Any,
    ) -> List[score_result.ScoreResult]:
        """
        Evaluate the output against all assertions.

        Args:
            input: All inputs that the agent received. Can be a string, dict, list,
                or any JSON-serializable structure containing the user query and metadata.
            output: All outputs from the agent. Can be a string, dict, list,
                or any JSON-serializable structure containing the response and metadata.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            List[ScoreResult]: A list of ScoreResult objects, one per assertion.
                Each result has:
                - name: The assertion text
                - value: True if passed, False if failed
                - reason: Explanation from the judge
        """
        try:
            return self._generate_and_parse(input=input, output=output)
        except LLMJudgeParseError as e:
            LOGGER.warning(
                "LLMJudge scoring failed after retries: %s", e, exc_info=True
            )
            return e.results

    @_RETRY_POLICY
    def _generate_and_parse(
        self,
        input: Any,
        output: Any,
    ) -> List[score_result.ScoreResult]:
        schema = parsers.ResponseSchema(self._assertions)
        llm_query = _generate_prompt(
            input=input,
            output=output,
            assertions_text=schema.format_assertions(),
        )
        model_output = self._model.generate_string(
            input=llm_query, response_format=schema.response_format
        )
        return schema.parse(model_output)

    async def ascore(
        self,
        input: Any,
        output: Any,
        **ignored_kwargs: Any,
    ) -> List[score_result.ScoreResult]:
        """
        Asynchronously evaluate the output against all assertions.

        Args:
            input: All inputs that the agent received. Can be a string, dict, list,
                or any JSON-serializable structure containing the user query and metadata.
            output: All outputs from the agent. Can be a string, dict, list,
                or any JSON-serializable structure containing the response and metadata.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            List[ScoreResult]: A list of ScoreResult objects, one per assertion.
        """
        try:
            return await self._agenerate_and_parse(input=input, output=output)
        except LLMJudgeParseError as e:
            LOGGER.warning(
                "LLMJudge async scoring failed after retries: %s", e, exc_info=True
            )
            return e.results

    @_RETRY_POLICY
    async def _agenerate_and_parse(
        self,
        input: Any,
        output: Any,
    ) -> List[score_result.ScoreResult]:
        schema = parsers.ResponseSchema(self._assertions)
        llm_query = _generate_prompt(
            input=input,
            output=output,
            assertions_text=schema.format_assertions(),
        )
        model_output = await self._model.agenerate_string(
            input=llm_query, response_format=schema.response_format
        )
        return schema.parse(model_output)

    def to_config(self) -> llm_judge_config.LLMJudgeConfig:
        """
        Serialize the evaluator configuration to a Pydantic model compatible
        with the backend's online evaluation system.

        The returned structure mirrors the `LlmAsJudgeCode` JSON format used
        in the `automation_rule_evaluators.code` column.

        Returns:
            LLMJudgeConfig: A Pydantic model containing:
                - model: Model configuration (name, temperature, seed)
                - messages: Prompt template messages
                - variables: Variable mappings for trace field extraction
                - schema: Output schema definitions for each assertion

        Example:
            >>> evaluator = LLMJudge(
            ...     assertions=["Response is accurate"],
            ...     temperature=0.0,
            ... )
            >>> config = evaluator.to_config()
        """
        model_config = llm_judge_config.LLMJudgeModelConfig(
            name=None,
            temperature=self._temperature,
            seed=self._seed,
            custom_parameters={"reasoning_effort": self._reasoning_effort},
        )

        messages = [
            llm_judge_config.LLMJudgeMessage(
                role="SYSTEM",
                content=LLM_JUDGE_SYSTEM_PROMPT,
            ),
            llm_judge_config.LLMJudgeMessage(
                role="USER",
                content=LLM_JUDGE_USER_TEMPLATE,
            ),
        ]

        variables = {
            "input": "input",
            "output": "output",
        }

        schema_items = [
            llm_judge_config.LLMJudgeSchemaItem(
                name=assertion,
                type="BOOLEAN",
                description=assertion,
            )
            for assertion in self._assertions
        ]

        return llm_judge_config.LLMJudgeConfig(
            name=self.name,
            model=model_config,
            messages=messages,
            variables=variables,
            schema=schema_items,
        )

    @classmethod
    def from_config(
        cls,
        config: llm_judge_config.LLMJudgeConfig,
        track: bool = True,
        project_name: Optional[str] = None,
        init_kwargs: Optional[Dict[str, Any]] = None,
    ) -> "LLMJudge":
        """
        Create an LLMJudge instance from a configuration.

        This method allows reconstructing an LLMJudge from a serialized config,
        typically retrieved from the backend's online evaluation system.

        Args:
            config: LLMJudgeConfig with model, messages, variables, and schema.
            track: Whether to track the evaluator. Defaults to True.
            project_name: Optional project name for tracking.
            init_kwargs: Optional dictionary to override __init__ parameters.
                Supported: 'model' to specify the model name.

        Returns:
            LLMJudge: A new instance configured according to the provided config.

        Example:
            >>> config = llm_judge_config.LLMJudgeConfig(
            ...     model=llm_judge_config.LLMJudgeModelConfig(temperature=0.0),
            ...     messages=[llm_judge_config.LLMJudgeMessage(role="USER", content="...")],
            ...     variables={"input": "input", "output": "output"},
            ...     schema=[llm_judge_config.LLMJudgeSchemaItem(name="accurate", type="BOOLEAN", description="Response is accurate")],
            ... )
            >>> evaluator = LLMJudge.from_config(config, init_kwargs={"model": "gpt-4o"})
        """
        assertion_texts = [item.description for item in config.schema_]

        init_kwargs = init_kwargs or {}
        model = init_kwargs.get("model")

        custom = config.model.custom_parameters or {}
        reasoning_effort = custom.get("reasoning_effort")

        return cls(
            assertions=assertion_texts,
            name=config.name,
            model=model,
            track=track,
            project_name=project_name,
            seed=config.model.seed,
            temperature=config.model.temperature,
            reasoning_effort=reasoning_effort,
        )



import logging
import warnings
from typing import Any


_FILTERS_INSTALLED = False


def add_warning_filters() -> None:
    global _FILTERS_INSTALLED
    if _FILTERS_INSTALLED:
        return

    warnings.filterwarnings("ignore", message="coroutine '.*' was never awaited")
    warnings.filterwarnings(
        "ignore",
        message="Enable tracemalloc to get the object allocation traceback",
    )
    warnings.filterwarnings(
        "ignore",
        message=r"(?s)The secret `HF_TOKEN` does not exist in your Colab secrets\..*",
        category=UserWarning,
        module=r"huggingface_hub\.utils\._auth",
    )
    warnings.filterwarnings(
        "ignore",
        category=UserWarning,
        message=r"You are sending unauthenticated requests.*",
    )
    warnings.filterwarnings(
        "ignore",
        category=UserWarning,
        message=r"Some weights of the model checkpoint.*",
    )
    warnings.filterwarnings(
        "ignore",
        category=UserWarning,
        message=r"Some weights of the model checkpoint were not used when initializing.*",
    )
    warnings.filterwarnings(
        "ignore",
        category=UserWarning,
        message=r"Unused arguments passed to model\.forward.*",
    )

    class NoEventLoopFilterLiteLLM(logging.Filter):
        def filter(self, record: Any) -> bool:
            return (
                "Asynchronous processing not initialized as we are not running in an async context"
                not in record.getMessage()
            )

    # Add filter to multiple possible loggers
    lite_logger = logging.getLogger("LiteLLM")
    has_filter = any(
        isinstance(f, NoEventLoopFilterLiteLLM) for f in lite_logger.filters
    )
    if not has_filter:
        lite_logger.addFilter(NoEventLoopFilterLiteLLM())

    import litellm

    litellm.suppress_debug_info = True

    _FILTERS_INSTALLED = True



"""Runner activation — called after the user's module loads to start the in-process loop."""

import logging
import os
import threading

from rich.console import Console
from rich.text import Text

from .. import Opik
from ..rest_api.types.agent import Agent
from ..rest_api.types.param import Param
from . import prefixed_output, registry
from .in_process_loop import InProcessRunnerLoop

LOGGER = logging.getLogger(__name__)

_started = False
_lock = threading.Lock()


def activate_runner() -> None:
    """Start the runner loop in a background thread (non-blocking)."""
    if os.environ.get("OPIK_RUNNER_MODE") != "true":
        return

    global _started
    with _lock:
        if _started:
            return
        _started = True

    t = threading.Thread(target=_run, daemon=True)
    t.start()


def _run() -> None:
    runner_id = os.environ.get("OPIK_RUNNER_ID", "")
    project_name = os.environ.get("OPIK_PROJECT_NAME", "")

    if not runner_id:
        LOGGER.error("OPIK_RUNNER_ID not set, cannot activate runner")
        return

    _print_banner(runner_id, project_name)

    prefixed_output.install()

    client = Opik(_show_misconfiguration_message=False)
    api = client.rest_client

    def _to_payload(entry: dict) -> dict:
        return Agent(
            description=entry.get("docstring", ""),
            language="python",
            params=[Param(name=p.name, type=p.type) for p in entry.get("params", [])],
            timeout=0,
        ).dict()

    def _sync_agent(name: str) -> None:
        entry = registry.get_all().get(name)
        if entry is None:
            return
        try:
            api.runners.register_agents(runner_id, request={name: _to_payload(entry)})
        except Exception:
            LOGGER.debug("Failed to register agent '%s'", name, exc_info=True)

    registry.on_register(_sync_agent)

    entrypoints = registry.get_all()
    if entrypoints:
        api.runners.register_agents(
            runner_id,
            request={name: _to_payload(entry) for name, entry in entrypoints.items()},
        )

    LOGGER.info("Runner activated")

    shutdown_event = threading.Event()
    loop = InProcessRunnerLoop(api, runner_id, shutdown_event)

    try:
        loop.run()
    finally:
        client.end()


def _print_banner(runner_id: str, project_name: str) -> None:
    console = Console()

    info = Text()
    info.append("   ")
    info.append("\u2800\u20dd", style="rgb(224,62,45)")
    info.append("opik  ", style="bold")
    info.append(f"runner: {runner_id}", style="dim")
    if project_name:
        info.append(f"  project: {project_name}", style="dim")
    console.print(info)
    console.print()
